
import csv
import zipfile
from Acquisition import aq_base
from Acquisition import aq_parent
from Products.CMFCore.utils import getToolByName
from bs4 import BeautifulSoup
from cStringIO import StringIO
from collective.contentreports.tool import register_report
from email import encoders
from email.mime.base import MIMEBase
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from plone.app.redirector.interfaces import IRedirectionStorage
from zope.component import getUtility
from zope.component.hooks import getSite

MESSAGE_BODY_FORMAT = (
    '\nHello,\n\n'
    'Attached is a %(report_name)s requested by %(requester)s on %(timestamp)s. '
    'The report includes information for the following area of the CNM '
    'website: %(url)s\n\n'
    'If you have any questions, contact the CNM Web Team at '
    'webstrategy@cnm.edu.'
)

IGNORED_LOCAL_ROLES = ('Owner', )
IGNORED_GLOBAL_ROLES = ('Member', 'Authenticated',)
IGNORED_GROUPS = ('AuthenticatedUsers',)


class BaseReport(object):

    def mail_report(self, report_request, report, mimetype, filename):
        site = getSite()
        buffer = StringIO()
        zfile = zipfile.ZipFile(buffer, 'w', zipfile.ZIP_DEFLATED)
        zfile.writestr(filename, report)
        zfile.close()
        mimetype = 'application/zip'
        parts = filename.split('.')
        parts[-1] = 'zip'
        filename = '.'.join(parts)
        mhost = getToolByName(site, 'MailHost')
        msg = MIMEMultipart()
        msg['Subject'] = 'CNM Website Report'
        requester_name = report_request['requester_name']
        if not requester_name:
            requester_name = report_request['requester_id']
        msg_body = MESSAGE_BODY_FORMAT % {
            'report_name': self.title,
            'url': report_request['url'],
            'requester': requester_name,
            'timestamp': report_request['timestamp'].strftime('%m/%d/%Y %H:%M')
        }
        msg.attach(MIMEText(msg_body))
        maintype, subtype = mimetype.split('/', 1)
        if maintype == 'text':
            attachment = MIMEText(report, _subtype=subtype)
        else:
            attachment = MIMEBase(maintype, subtype)
            attachment.set_payload(buffer.getvalue())
            encoders.encode_base64(attachment)
        attachment.add_header('Content-Disposition', 'attachment',
                              filename=filename)
        msg.attach(attachment)
        mto = report_request['recipients']
        mfrom = site.getProperty('email_from_address')
        if not mfrom:
            mfrom = report_request.get('requester_email')
        if not mfrom:
            mfrom = mto[0]
        mhost.send(msg, mto=mto, mfrom=mfrom)


def inherited(context):
    if getattr(aq_base(context), '__ac_local_roles_block__', None):
        return False
    return True


def get_inherited_roles(context):

    if not inherited(context):
        return []

    portal = getToolByName(context, 'portal_url').getPortalObject()
    result = []
    cont = True
    if portal != context:
        parent = aq_parent(context)
        while cont:
            if not getattr(parent, 'acl_users', False):
                break
            userroles = parent.acl_users._getLocalRolesForDisplay(parent)
            for user, roles, role_type, name in userroles:
                # Find user in result
                found = 0
                for user2, roles2, type2, name2 in result:
                    if user2 == user:
                        # Check which roles must be added to roles2
                        for role in roles:
                            if not role in roles2:
                                roles2.append(role)
                        found = 1
                        break
                if found == 0:
                    # Add it to result and make sure roles is a list so
                    # we may append and not overwrite the loop variable
                    result.append([user, list(roles), role_type, name])
            if parent == portal:
                cont = False
            elif not inherited(parent):
                # Role acquired check here
                cont = False
            else:
                parent = aq_parent(parent)

    # Tuplize all inner roles
    for pos in range(len(result)-1, -1, -1):
        result[pos][1] = tuple(result[pos][1])
        result[pos] = tuple(result[pos])

    return tuple(result)


def get_shared_roles(obj):
    acl_users = getToolByName(obj, 'acl_users')
    mapping = dict()
    for name, roles, type, id in acl_users._getLocalRolesForDisplay(obj):
        roles = [r for r in roles if r not in IGNORED_LOCAL_ROLES]
        if not roles:
            continue
        mapping[id] = dict(
            id=id,
            type=type,
            local=roles,
            name=name,
            inherited=list(),
        )

    for name, roles, type, id in get_inherited_roles(obj):
        roles = [r for r in roles if r not in IGNORED_LOCAL_ROLES]
        if not roles:
            continue
        if id in mapping:
            mapping[id]['inherited'] = roles
        else:
            mapping[id] = dict(
                id=id,
                type=type,
                local=list(),
                name=name,
                inherited=roles,
            )
    return mapping


class WebsiteInventoryReport(BaseReport):

    name = 'website_inventory_report'

    title = u'Website Inventory'

    description = u'A csv file of all of the content in the selected location.'

    def run_report(self, report_request):
        site = getSite()
        catalog = getToolByName(site, 'portal_catalog')
        alias_storage = getUtility(IRedirectionStorage)
        buffer = StringIO()
        columns = ('path', 'title', 'url', 'type', 'state', 'excluded',
                   'created', 'published', 'modified', 'aliases', 'size',
                   'groups', 'users')
        writer = csv.DictWriter(buffer, columns)
        if hasattr(writer, 'writeheader'):
            writer.writeheader()
        else:
            header_row = dict()
            for column in columns:
                header_row[column] = column
            writer.writerow(header_row)
        for brain in catalog.unrestrictedSearchResults(
                path=report_request['scope']):
            data = dict()
            try:
                obj = brain.getObject()
            except AttributeError:
                continue
            path = brain.getPath()
            data['path'] = path
            data['url'] = brain.getURL()
            data['type'] = brain.Type
            data['title'] = brain.Title
            data['state'] = brain.review_state
            data['excluded'] = brain.exclude_from_nav and 'yes' or 'no'
            data['created'] = brain.CreationDate
            data['published'] = brain.EffectiveDate
            data['modified'] = brain.ModificationDate
            data['aliases'] = ';'.join(alias_storage.redirects(path))
            data['size'] = brain.getObjSize
            roles_mapping = get_shared_roles(obj)
            groups = list()
            users = list()
            for id, record in roles_mapping.items():
                if record['type'] == 'group':
                    groups.append(id)
                else:
                    users.append(id)
            groups.sort()
            users.sort()
            data['groups'] = ';'.join(groups)
            data['users'] = ';'.join(users)
            writer.writerow(data)
        self.mail_report(report_request, buffer.getvalue(), 'text/csv',
                         'website_inventory_report.csv')

register_report(WebsiteInventoryReport())


def add_label(soup, parent, label):
    tag = soup.new_tag('h3')
    tag.append(label)
    parent.append(tag)


def get_member_data(portal_groups, member):
    id = member.getId()
    data = dict()
    fullname = member.getProperty('fullname')
    if not fullname:
        fullname = id
    data['fullname'] = fullname
    data['email'] = member.getProperty('email')
    data['global_roles'] = [r for r in member.getRoles()
                            if r not in IGNORED_GLOBAL_ROLES]
    groups = portal_groups.getGroupsForPrincipal(member.getUser())
    groups = [g for g in groups if g not in IGNORED_GROUPS]
    data['groups'] = groups
    data['sharing'] = list()
    return data


class UserReport(BaseReport):

    name = 'user_report'

    title = u'User Report'

    description = u'A report of the site users'

    def run_report(self, report_request):
        site = getSite()
        portal_groups = getToolByName(site, 'portal_groups')
        portal_membership = getToolByName(site, 'portal_membership')
        acl_users = getToolByName(site, 'acl_users')
        members = dict()

        # for group in portal_groups.listGroups():
        #     for member in group.getGroupMembers():
        #         memberid = member.getId()
        #         if memberid not in members:
        #             members[memberid] = get_member_data(portal_groups, member)


        role_manager = acl_users['portal_role_manager']
        for role in role_manager.listRoleIds():
            if role in IGNORED_GLOBAL_ROLES:
                pass
            for memberid, title in role_manager.listAssignedPrincipals(role):
                if memberid in members:
                    continue
                member = portal_membership.getMemberById(memberid)
                if member is not None:
                    members[memberid] = get_member_data(portal_groups, member)
                else:
                    group = portal_groups.getGroupById(memberid)
                    if group is not None:
                        for group_memberid in group.getAllGroupMemberIds():
                            if group_memberid in members:
                                continue
                            group_member = portal_membership.getMemberById(
                                group_memberid)
                            if group_member is None:
                                continue
                            members[group_memberid] = get_member_data(
                                portal_groups, group_member)

        catalog = getToolByName(site, 'portal_catalog')
        for brain in catalog.unrestrictedSearchResults(
                path=report_request['scope']):
            try:
                obj = brain.getObject()
            except AttributeError:
                continue
            roles_mapping = get_shared_roles(obj)
            for memberid, roles_record in roles_mapping.items():
                if roles_record['type'] == 'group':
                    group = portal_groups.getGroupById(memberid)
                    if group is not None:
                        for group_memberid in group.getAllGroupMemberIds():
                            if group_memberid not in members:
                                group_member = portal_membership.getMemberById(
                                    group_memberid)
                                if group_member is None:
                                    continue
                                members[group_memberid] = get_member_data(
                                    portal_groups, group_member)
                            members[group_memberid]['sharing'].append(
                                brain.getURL()
                            )
                else:
                    if memberid not in members:
                        member = portal_membership.getMemberById(memberid)
                        if member is None:
                            continue
                        members[memberid] = get_member_data(portal_groups,
                                                            member)
                    members[memberid]['sharing'].append(brain.getURL())

        ids = members.keys()
        ids.sort()

        soup = BeautifulSoup('<html><body></body></html>')

        for memberid in ids:
            record = members[memberid]
            utag = soup.new_tag('div')
            utag['class'] = 'user'
            soup.body.append(utag)
            add_label(soup, utag, 'Name')
            tag = soup.new_tag('div')
            tag['class'] = 'user_name'
            tag.append(record['fullname'])
            utag.append(tag)
            add_label(soup, utag, 'Email')
            tag = soup.new_tag('div')
            tag['class'] = 'user_email'
            tag.append(record['email'])
            utag.append(tag)
            add_label(soup, utag, 'Global Roles')
            roles_tag = soup.new_tag('div')
            roles_tag['class'] = 'global_roles'
            utag.append(roles_tag)
            for role in record['global_roles']:
                tag = soup.new_tag('div')
                tag['class'] = 'role'
                tag.append(role)
                roles_tag.append(tag)
            add_label(soup, utag, 'Groups')
            groups_tag = soup.new_tag('div')
            groups_tag['class'] = 'groups'
            utag.append(groups_tag)
            for group in record['groups']:
                group_tag = soup.new_tag('div')
                group_tag['class'] = 'group'
                group_tag.append(group)
                groups_tag.append(group_tag)
            add_label(soup, utag, 'Sharing')
            sharing_tag = soup.new_tag('div')
            utag.append(sharing_tag)
            for url in record['sharing']:
                tag = soup.new_tag('div')
                tag['class'] = 'url'
                tag.append(url)
                sharing_tag.append(tag)
            soup.body.append(soup.new_tag('hr'))
        self.mail_report(report_request, soup.prettify(), 'text/html',
                         'users_report.html')


register_report(UserReport())


class GroupReport(BaseReport):

    name = 'group_report'

    title = u'Group Report'

    description = u'A report of the site groups'

    def run_report(self, report_request):
        site = getSite()
        portal_groups = getToolByName(site, 'portal_groups')
        groups = dict()
        for group in portal_groups.listGroups():
            groups[group.getGroupId()] = data = dict()
            data['name'] = group.getGroupName()
            data['title'] = group.getProperty('title')
            data['global_roles'] = [r for r in group.getRoles()
                                    if r not in IGNORED_GLOBAL_ROLES]
            members = list()
            for member in group.getGroupMembers():
                name = member.getProperty('fullname')
                if not name:
                    name = member.getId()
                members.append(dict(name=name,
                                    email=member.getProperty('email')))
            data['members'] = members
            data['sharing'] = list()

        catalog = getToolByName(site, 'portal_catalog')
        for brain in catalog.unrestrictedSearchResults(
                path=report_request['scope']):
            try:
                obj = brain.getObject()
            except AttributeError:
                continue
            roles_mapping = get_shared_roles(obj)
            for id in roles_mapping:
                if id in groups:
                    groups[id]['sharing'].append((brain.getURL()))

        ids = groups.keys()
        ids.sort()

        soup = BeautifulSoup('<html><body></body></html>')
        for id in ids:
            record = groups[id]
            if not record['sharing'] and not record['global_roles']:
                continue
            gtag = soup.new_tag('div')
            gtag['class'] = 'group'
            soup.body.append(gtag)
            add_label(soup, gtag, 'Name')
            tag = soup.new_tag('div')
            tag['class'] = 'group_name'
            tag.append(record['name'])
            gtag.append(tag)
            add_label(soup, gtag, 'Title')
            tag = soup.new_tag('div')
            tag['class'] = 'group_title'
            tag.append(record['title'])
            gtag.append(tag)
            add_label(soup, gtag, 'Members')
            members_tag = soup.new_tag('div')
            members_tag['class'] = 'members'
            gtag.append(members_tag)
            for member in record['members']:
                member_tag = soup.new_tag('div')
                member_tag['class'] = 'member'
                members_tag.append(member_tag)
                tag = soup.new_tag('div')
                tag['class'] = 'member_name'
                tag.append(member['name'])
                member_tag.append(tag)
                tag = soup.new_tag('div')
                tag['class'] = 'member_email'
                tag.append(member['email'])
                member_tag.append(tag)
            add_label(soup, gtag, 'Sharing')
            sharing_tag = soup.new_tag('div')
            gtag.append(sharing_tag)
            for url in record['sharing']:
                tag = soup.new_tag('div')
                tag['class'] = 'url'
                tag.append(url)
                sharing_tag.append(tag)
            soup.body.append(soup.new_tag('hr'))
        self.mail_report(report_request, soup.prettify(), 'text/html',
                         'groups_report.html')

register_report(GroupReport())
